import datetime
from . import Plugin, tool


class Clock(Plugin):
    @tool
    def get_current_utc_time(self):
        """Get the current UTC time in ISO format"""
        return datetime.datetime.now(datetime.timezone.utc).isoformat()



import asyncio
from dataclasses import dataclass, is_dataclass
import enum
import inspect
import json
import logging
import os
import types
from typing import (
    Any,
    Callable,
    TypeVar,
    Coroutine,
    Optional,
    Union,
    cast,
    overload,
    get_args,
    get_origin,
    Annotated,
    TYPE_CHECKING,
)
from pydantic import BaseModel, TypeAdapter
from openai.lib._pydantic import _ensure_strict_json_schema  # type: ignore
from PIL.Image import Image
import base64
from io import BytesIO
import inspect

from agentia.spec import UserMessage, MessagePartText, MessagePartFile, SystemMessage
from agentia.spec.chat import Message

if TYPE_CHECKING:
    from agentia.tools.tools import Tools


@dataclass
class ImageUrl:
    """
    A special class for magic functions to represent an image URL argument.
    """

    url: str


def _is_image_type(t: type) -> bool:
    if t == ImageUrl | Image:
        return True
    if not inspect.isclass(t):
        return False
    return issubclass(t, Image) or issubclass(t, ImageUrl)


class ToolFuncParam:
    def __init__(self, param: inspect.Parameter, fname: str, is_magic=False) -> None:
        self.is_magic = is_magic
        self.func_name = fname
        self.param = param
        self.name = param.name
        t, r = self.__get_type()
        self.type = t
        self.required = r
        self.description = self.__get_desc()
        self.default = self.__get_default()
        self.is_self = self.name == "self"
        self.schema = self.__get_json_schema()

    def __get_json_schema(self) -> dict[str, Any] | None:
        if self.is_magic and _is_image_type(self.type):
            return None

        if inspect.isclass(self.type) and issubclass(self.type, BaseModel):
            schema = self.type.model_json_schema()
        else:
            schema = TypeAdapter(self.type).json_schema()
        if self.description:
            if "description" in schema:
                schema["description"] += self.description
            else:
                schema["description"] = self.description
        return _ensure_strict_json_schema(schema, path=(), root=schema)

    def __get_desc(self) -> str | None:
        # if type has a docstring, add to description
        type_desc = None
        if is_dataclass(self.type) or isinstance(self.type, BaseModel):
            if desc := self.type.__doc__:
                type_desc = desc
        meta = (
            get_args(self.param.annotation)[1]
            if get_origin(self.param.annotation) == Annotated
            else None
        )
        if meta is not None and isinstance(meta, str):
            if type_desc:
                return type_desc + "\n\n" + meta
            return meta
        if type_desc:
            return type_desc
        return None

    def __get_default(self) -> Any | None:
        if self.param.default == inspect.Parameter.empty:
            return None
        else:
            return self.param.default

    def __get_type(self) -> tuple[type, bool]:
        t = self.param.annotation
        t = t if get_origin(t) != Annotated else get_args(t)[0]
        if t == inspect.Parameter.empty:
            t = str  # the default type is string
        # Get parameter optionality
        param_t_is_opt = False
        is_optional = lambda x: (
            (get_origin(x) is Union or get_origin(x) is types.UnionType)
            and len(get_args(x)) == 2
            and type(None) in get_args(x)
        )
        if is_optional(t):
            t, param_t_is_opt = get_args(t)[0], True
        param_default_is_empty = self.param.default == inspect.Parameter.empty
        required = not param_t_is_opt and param_default_is_empty
        # Get parameter type
        assert not is_optional(t), "Optional types are not supported"
        return t, required


R = TypeVar("R", Coroutine[Any, Any, Optional[Any | str]], Optional[Any | str])


@overload
def tool(func: Callable[..., R], /) -> Callable[..., R]:
    """
    Decorator to mark a function as a tool for the agent.
    """
    ...


@overload
def tool(
    *,
    name: str | None = None,
    display_name: str | None = None,
    description: str | None = None,
    metadata: Any | None = None,
) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """
    Decorator to mark a function as a tool for the agent.
    """
    ...


def tool(
    func: Callable[..., R] | None = None,
    /,
    *,
    name: str | None = None,
    display_name: str | None = None,
    description: str | None = None,
    metadata: Any | None = None,
) -> Callable[..., R] | Callable[[Callable[..., R]], Callable[..., R]]:
    """
    Decorator to mark a function as a tool for the agent.
    """

    def __tool_impl(callable: Callable[..., R]) -> Callable[..., R]:
        from agentia.tools.tools import (
            NAME_TAG,
            DISPLAY_NAME_TAG,
            DESCRIPTION_TAG,
            METADATA_TAG,
            IS_TOOL_TAG,
        )

        # store gpt function metadata to the callable object
        if isinstance(name, str):
            setattr(callable, NAME_TAG, name)
        if isinstance(display_name, str):
            setattr(callable, DISPLAY_NAME_TAG, display_name)
        if isinstance(description, str):
            setattr(callable, DESCRIPTION_TAG, description)
        if metadata is not None:
            setattr(callable, METADATA_TAG, metadata)
        setattr(callable, IS_TOOL_TAG, True)
        return callable

    if func is not None:
        return __tool_impl(func)

    return __tool_impl


F = TypeVar("F", bound=Callable)


def _gen_prompt(
    f: Callable,
    args: list[Any],
    kwargs: dict[str, Any],
    instructions: str | None,
) -> tuple[str, list[tuple[ToolFuncParam, Image | ImageUrl]]]:
    desc = "You need to do the following task and return the result in JSON:\n\n"
    desc += "TASK NAME: " + f.__name__ + "\n"
    if instructions:
        desc += "DESCRIPTION: \n" + instructions + "\n\n"
    elif doc := f.__doc__:
        desc += "DESCRIPTION: \n" + doc + "\n\n"
    else:
        logging.getLogger("agentia").warning(
            f"WARNING: Function {f.__name__} has no docstring. It's recommended to use the docstring to provide the agent with a description of the task."
        )
    sig = inspect.signature(f)
    images: list[tuple[ToolFuncParam, Image | ImageUrl]] = []
    if len(sig.parameters) > 0:
        params = [
            ToolFuncParam(p, f.__name__, is_magic=True)
            for p in sig.parameters.values()
            if p.name != "self"
        ]
        params_with_values: dict[str, tuple[ToolFuncParam, Any]] = {}
        for a in args:
            # get and remove the first positional argument from params
            param = None
            for p in params:
                if p.param.kind in (
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.POSITIONAL_ONLY,
                ):
                    params.remove(p)
                    param = p
                    break
            if param is None:
                raise ValueError("Too many positional arguments")
            else:
                params_with_values[param.name] = (param, a)
        for k, v in kwargs.items():
            # get and remove the first keyword argument from params
            param = None
            for p in params:
                if p.param.name != k:
                    continue
                if p.param.kind in (
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ):
                    params.remove(p)
                    param = p
                    break
            if param is None:
                raise ValueError("Unexpected keyword argument: " + k)
            else:
                params_with_values[param.name] = (param, v)
        for p in params:
            if p.name in params_with_values:
                continue
            if p.required:
                raise ValueError("Missing argument: " + p.name)
            params_with_values[p.name] = (p, p.default)
        desc += "USER'S INPUTS:\n\n"
        for p, value in params_with_values.values():
            if not p.schema:
                continue
            pdesc = " * " + p.name + ": "
            if isinstance(value, BaseModel):
                value_str = value.model_dump_json()
            else:
                value_str = json.dumps(value)
            if p.description or (p.schema and "enum" in p.schema):
                pdesc += "\n"
                if s := p.description:
                    pdesc += f"    * description: {s}\n"
                if s := (p.schema or {}).get("enum"):
                    pdesc += f"    * possible values: {', '.join(s)}\n"
                pdesc += "    * value: " + value_str + "\n"
            else:
                pdesc += value_str + "\n"
            desc += pdesc
        for p, value in params_with_values.values():
            if _is_image_type(p.type):
                value = cast(Image | ImageUrl, value)
                images.append((p, value))
    return desc, images


@overload
def magic(func: F, /) -> F: ...


@overload
def magic(
    *,
    model: str | None = None,
    instructions: str | None = None,
    tools: Optional["Tools"] = None,
) -> Callable[[F], F]: ...


def _is_supported_magic_return_type(t: type) -> bool:
    if t in (int, float, str, bool, type(None)):
        return True
    if get_origin(t) is Union or get_origin(t) is types.UnionType:
        args = get_args(t)
        if len(args) == 2 and type(None) in args:
            other_type = args[0] if args[1] is type(None) else args[1]
            return _is_supported_magic_return_type(other_type)
    # Enum
    if inspect.isclass(t) and issubclass(t, enum.Enum):
        return True
    # tuple, list, dict
    if get_origin(t) in (list, tuple, dict):
        args = get_args(t)
        for arg in args:
            if not _is_supported_magic_return_type(arg):
                return False
        return True
    # Pydantic BaseModel
    if inspect.isclass(t):
        return issubclass(t, BaseModel)
    return False


def magic(
    func: F | None = None,
    /,
    *,
    model: str | None = None,
    instructions: str | None = None,
    tools: Optional["Tools"] = None,
) -> F | Callable[[F], F]:
    """
    Decorator to mark a function as an agent for the agent.
    """
    if not model:
        model = os.getenv("AGENTIA_DEFAULT_MODEL", "openai/gpt-5-mini")

    def __magic_impl(callable: F) -> F:
        async def __func_impl(*args: Any, **kwargs: Any):
            from agentia.llm import LLM

            prompt, images = _gen_prompt(callable, list(args), kwargs, instructions)

            llm = LLM(model=model or "openai/gpt-5-mini")
            return_type = inspect.signature(callable).return_annotation
            if isinstance(return_type, inspect._empty):
                return_type = str
            if not _is_supported_magic_return_type(return_type):
                raise ValueError(
                    f"Unsupported return type: {return_type} in magic function {callable.__name__}."
                )

            messages: list[Message] = [SystemMessage(content=prompt)]
            for i, (p, image) in enumerate(images):
                content_type = "image/png"
                if isinstance(image, ImageUrl):
                    url = image.url
                    url_lower = url.lower()
                    if url_lower.endswith((".jpg", ".jpeg")):
                        content_type = "image/jpeg"
                    elif url_lower.endswith(".png"):
                        content_type = "image/png"
                    elif url_lower.endswith(".gif"):
                        content_type = "image/gif"
                    elif url_lower.endswith(".bmp"):
                        content_type = "image/bmp"
                    elif url_lower.endswith(".webp"):
                        content_type = "image/webp"
                    elif url_lower.endswith(".ico"):
                        content_type = "image/ico"

                else:
                    match image.format:
                        case "JPEG":
                            content_type = "image/jpeg"
                        case "PNG":
                            content_type = "image/png"
                        case "GIF":
                            content_type = "image/gif"
                        case "BMP":
                            content_type = "image/bmp"
                        case "WEBP":
                            content_type = "image/webp"
                        case "ICO":
                            content_type = "image/ico"
                        case _:
                            raise ValueError(
                                f"Unsupported image format: {image.format}"
                            )
                    buffered = BytesIO()
                    buffered.name = f"image.{content_type}"
                    image.save(buffered, format=image.format)
                    img_data = base64.b64encode(buffered.getvalue()).decode("utf-8")
                    url = f"data:image/{content_type};base64,{img_data}"
                s = f"Image Argument #{i}: {p.name}"
                if p.description:
                    s += f"\nDescription: {p.description})"
                messages.append(
                    UserMessage(
                        content=[
                            MessagePartText(text=s),
                            MessagePartFile(data=url, media_type=content_type),
                        ],
                        role="user",
                    )
                )
            r = llm.generate(messages, options={"tools": tools})
            await r
            messages.extend(r.new_messages)
            messages.append(
                UserMessage(
                    content=[MessagePartText(text="Output the result in JSON format")]
                )
            )
            result = await llm.generate_object(messages, type=return_type)
            return result

        if not inspect.iscoroutinefunction(callable):

            def __func_impl_sync(*args: Any, **kwargs: Any):
                return asyncio.run(__func_impl(*args, **kwargs))

            return __func_impl_sync  # type: ignore

        return __func_impl  # type: ignore

    if func is not None:
        return __magic_impl(func)

    return __magic_impl



from ._openai_api import OpenAIAPIProvider
import os


class OpenRouter(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        api_key = api_key or os.environ.get("OPENROUTER_API_KEY")
        if not api_key:
            raise ValueError("OPENROUTER_API_KEY environment variable not set")
        base_url = "https://openrouter.ai/api/v1"
        super().__init__(
            provider="openrouter", model=model, api_key=api_key, base_url=base_url
        )
        m = os.getenv("AGENTIA_OPENROUTER_MODALITIES", None)
        if m:
            self.extra_body["modalities"] = [
                x.strip().lower() for x in m.split(",") if x.strip()
            ]
        t = os.getenv("AGENTIA_OPENROUTER_TRANSFORMS", None)
        if t and t != "0" and t.lower() != "false":
            self.extra_body["transforms"] = ["middle-out"]
        if p := os.getenv("OPENROUTER_PROVIDERS", None):
            self.extra_body["provider"] = {
                "order": [x.strip().lower() for x in p.split(",") if x.strip()],
            }



from typing import AsyncGenerator, Literal, Optional, TYPE_CHECKING

from agentia.llm.completion import Listeners, async_gen_to_sync
from agentia.spec import *

if TYPE_CHECKING:
    from agentia.agent import Agent


class TextStream:
    def __init__(self, gen: AsyncGenerator[str, None]):
        async def _gen():
            async for part in gen:
                self.__all += part
                yield part

        self.__all = ""
        self.__gen = _gen()

    def __aiter__(self) -> AsyncGenerator[str, None]:
        return self.__gen

    async def __wait_for_completion(self) -> str:
        async for _ in self.__gen:
            pass
        return self.__all

    def __await__(self):
        return self.__wait_for_completion().__await__()

    def __iter__(self):
        return async_gen_to_sync(self.__gen)


class ReasoningStream(TextStream):
    type: Literal["reasoning-stream"] = "reasoning-stream"


class MessageStream(TextStream):
    type: Literal["message-stream"] = "message-stream"


class ChatCompletionStreamBase:
    def __init__(self, agent: Optional["Agent"]):
        self.usage = Usage()
        self.finish_reason: FinishReason | None = None
        self.new_messages: list[Message] = []
        self.on_finish = Listeners()
        self.on_new_message = Listeners()
        self.agent = agent

    def _on_finish(self):
        self.on_finish.emit()
        if self.agent:
            self.agent.emit("finish")

    def add_new_message(self, msg: Message):
        self.new_messages.append(msg)
        self.on_new_message.emit(msg)


class ChatCompletionStream(ChatCompletionStreamBase):
    def __init__(self, gen: AsyncGenerator[StreamPart, None], agent: Optional["Agent"]):
        super().__init__(agent)

        async def __gen() -> (
            AsyncGenerator[
                ReasoningStream | MessageStream | ToolCall | ToolResult, None
            ]
        ):
            async for part in gen:
                if isinstance(part, (ToolCall, ToolResult)):
                    yield part
                elif isinstance(part, StreamPartTextStart):

                    async def gen2():
                        async for part2 in gen:
                            if isinstance(part2, StreamPartTextDelta):
                                yield part2.delta
                            elif isinstance(part2, StreamPartTextEnd):
                                break

                    yield MessageStream(gen2())
                elif isinstance(part, StreamPartReasoningStart):

                    async def gen2():
                        async for part2 in gen:
                            if isinstance(part2, StreamPartReasoningDelta):
                                yield part2.delta
                            elif isinstance(part2, StreamPartReasoningEnd):
                                break

                    yield ReasoningStream(gen2())
            self._on_finish()

        self.__gen = __gen()

    def __aiter__(
        self,
    ) -> AsyncGenerator[
        ReasoningStream | MessageStream | ToolCall | ToolResult | Annotation, None
    ]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for item in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()

    def __iter__(self):
        return async_gen_to_sync(self.__gen)


class ChatCompletionEvents(ChatCompletionStreamBase):
    def __init__(self, gen: AsyncGenerator[StreamPart, None], agent: Optional["Agent"]):
        super().__init__(agent)

        async def __gen():
            async for part in gen:
                yield part
            self._on_finish()

        self.__gen = __gen()

    def __aiter__(self) -> AsyncGenerator[StreamPart, None]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for item in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()



from ._openai_api import OpenAIAPIProvider
import os


class Ollama(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        base_url = os.environ.get(
            "OLLAMA_BASE_URL",
            os.environ.get("OPENAI_BASE_URL", "http://localhost:11434/v1"),
        )
        api_key = os.environ.get("OLLAMA_API_KEY", "dummy")
        think = model.endswith(":think")
        if think:
            model = model[:-6]
        super().__init__(
            provider="ollama", model=model, api_key=api_key, base_url=base_url
        )
        self.extra_body["think"] = think



import os
from ..utils.decorators import tool
from agentia.tools import Plugin, PluginInitError


__all__ = [
    "Plugin",
    "PluginInitError",
    "tool",
]

if os.environ.get("AGENTIA_DISABLE_PLUGINS", "").lower() not in [
    "1",
    "true",
    "yes",
    "y",
]:
    try:
        from .calc import Calculator
        from .clock import Clock
        from .code_runner import CodeRunner
        from .memory import Memory
        from .search import Search
        from .web import Web
        from .skills import Skills
        from .skill_learner import SkillLearner

    except ImportError as e:
        pass
        # raise e



import os
from pathlib import Path
from typing import Sequence
from agentia import (
    Agent,
    MessagePartText,
    UserMessage,
    UserMessagePart,
    MessagePartFile,
)
from agentia.plugins.skills import Skill
from pydantic import BaseModel, Field
from tavily import AsyncTavilyClient
from dotenv import load_dotenv
import frontmatter

load_dotenv()


class SkillScript(BaseModel):
    name: str = Field(
        description="The name of the script (without the path), e.g. 'data_processing.py'"
    )
    content: str = Field(
        description="The content of the script as a string. Must be valid Python code. Inputs are from CLI arguments. Outputs are through stdout (using print) either in JSON or raw text format."
    )


class SkillReference(BaseModel):
    name: str = Field(description="The name of the reference document, with .md suffix")
    description: str = Field(description="A brief description of the skill")
    content: str = Field(
        description="The markdown content of the reference document as a string."
    )


class SkillInfo(BaseModel):
    description: str = Field(description="A brief description of the skill")
    readme: str = Field(
        description="""
        A detailed description of the skill, including its detailed descriptions, capabilities, and usage instructions (or experiences).
        This should be in markdown format that can be reused by other LLM agents to follow/apply the skill.

        # Scripts

        Use scripts to provide executable code that can be used to perform tasks.

        When providing scripts:
            In this README, **add a section** that lists each script path and how to use it, so that other agents can use them.
            In this README, please specify the expected CLI arguments, expected output, and provide examples.

        Script paths should be `scripts/<name>.py`

        # Reference documents

        Use reference docs to provide more detailed information about the skill, such as case studies, examples, detailed explanations, etc.
        This can help other agents better understand the skill and apply it more effectively.

        When providing reference docs:
            In this README, **add a section** that lists each reference doc path and their short description, so that other agents can load it when needed to strengthen the skill.

        Please provide a brief summary of the key points in the document and how it can be used to strengthen the skill.

        Reference document paths should be `references/<name>.md`
        """
    )
    scripts: list[SkillScript] = Field(
        default_factory=list,
        description="(Optional, only when applicable) A list of scripts provided by the skill, if any. Scripts are placed in the 'scripts' folder.",
    )
    reference_docs: list[SkillReference] = Field(
        default_factory=list,
        description="(Optional, only when applicable) A list of reference documents for the skill (as appendix or supplementary materials), if any.",
    )

    def save(self, name: str, out: Path) -> Skill:
        out.mkdir(parents=True, exist_ok=True)

        if any(out.iterdir()):
            raise ValueError(f"Output directory '{out}' is not empty")

        assert out.is_dir(), f"Output path '{out}' must be a directory"
        skills_md = frontmatter.dumps(
            frontmatter.Post(
                content=self.readme, name=name, description=self.description
            )
        )
        (out / "SKILL.md").write_text(skills_md, encoding="utf-8")
        if self.scripts:
            (out / "scripts").mkdir(exist_ok=True)
        for f in self.scripts:
            if not f.name.endswith(".py"):
                f.name += ".py"
            (out / "scripts" / f.name).write_text(f.content, encoding="utf-8")
        if self.reference_docs:
            (out / "references").mkdir(exist_ok=True)
        for r in self.reference_docs:
            if not r.name.endswith(".md"):
                r.name += ".md"
            doc = frontmatter.dumps(
                frontmatter.Post(
                    content=r.content, name=r.name, description=r.description
                )
            )
            (out / "references" / r.name).write_text(doc, encoding="utf-8")

        skill = Skill(
            path=out,
            name=name,
            description=self.description,
            instructions=self.readme,
            script_paths=[f.name for f in self.scripts],
            resource_paths=[f.name for f in self.reference_docs],
        )
        return skill


async def learn_skill_from_documents(
    docs: Sequence[Path | str], name: str, prompt: str, out: Path, model: str
) -> Skill:
    instructions = "You are a skill generator/learner. Your task is to learn a skill based on the given documents. Output the skill in JSON format\n\n"
    instructions += f"Skill Name: {name}\n\n"
    instructions += f"{prompt}"

    agent = Agent(model=model, instructions=instructions)

    parts: list[UserMessagePart] = []

    urls: list[str] = []
    files: list[Path] = []
    allowed_suffixes = [".pdf", ".md", ".markdown", ".txt"]
    for x in docs:
        if isinstance(x, str) and x.startswith(("http://", "https://")):
            urls.append(x)
        else:
            p = Path(x)
            if p.is_file():
                files.append(p)
            elif p.is_dir():
                for f in p.rglob("*"):
                    if f.is_file() and f.suffix.lower() in allowed_suffixes:
                        files.append(f)

    # retrieve content from URLs using tavily
    if urls:
        tavily_api_key = os.getenv("TAVILY_API_KEY")
        if not tavily_api_key:
            raise ValueError("TAVILY_API_KEY is required to fetch web content")
        tavily = AsyncTavilyClient(api_key=tavily_api_key)
        try:
            result = await tavily.extract(urls=urls, extract_depth="advanced")
            results = result.get("results", [])
            for result in results:
                content = result["raw_content"]
                url = result["url"]
                text = "# Web Content: " + url + "\n\n" + content
                parts.append(MessagePartText(text=text))
        except Exception as e:
            raise ValueError(f"Failed to fetch web content: {e}")

    # retrieve content from files
    for doc in files:
        content = doc.read_text()
        media_type = "text/plain"
        suffix = doc.suffix.lower()
        if suffix in [".pdf"]:
            media_type = "application/pdf"
            parts.append(
                MessagePartFile(filename=doc.name, data=content, media_type=media_type)
            )
        elif suffix in [".md", ".markdown", ".txt"]:
            text = "# Document: " + doc.name + "\n\n" + content
            parts.append(MessagePartText(text=text))
        else:
            raise ValueError(f"Unsupported file type: {suffix}")

    if not parts:
        raise ValueError("No valid documents provided to learn the skill")

    skill_info = await agent.generate_object(UserMessage(parts), SkillInfo)

    skill = skill_info.save(name=name, out=out)

    return skill



from pathlib import Path
from pydantic import BaseModel
import pydantic
from . import Plugin, tool
from typing import Annotated
from datetime import datetime
from filelock import FileLock


class MemoryRecord(BaseModel):
    timestamp: datetime
    content: str


type MemoryRecords = list[MemoryRecord]


class Memory(Plugin):
    def __init__(self, file: str | Path):
        super().__init__()
        file = Path(file)
        assert file.suffix == ".json", "Memory file must be a .json file"
        self.__memory_cache = Path(file)
        self.__records: MemoryRecords = []

    async def init(self):
        assert self.agent is not None, "Agent is required for MemoryPlugin"
        # Load existing memory records
        assert self.__memory_cache.exists()
        content = self.__memory_cache.read_text().strip()
        records = pydantic.TypeAdapter[MemoryRecords](MemoryRecords).validate_json(
            content
        )
        self.__records = records
        # Append to agent history
        self.agent.history.add_instructions(f"YOUR PREVIOUS MEMORY: \n{content}")

    @tool
    def remember(
        self,
        content: Annotated[str, "The content to remember. Keep it short and brief."],
    ):
        """Permanently remember something in your memory, as long as you think it's important or will be useful in the future. Use this to remember any important information whilst you are chatting with the user or fulfilling tasks."""
        with FileLock(str(self.__memory_cache) + ".lock"):
            record = MemoryRecord(timestamp=datetime.now(), content=content)
            self.__records.append(record)
            # save to file
            with open(self.__memory_cache) as f:
                f.write(
                    pydantic.TypeAdapter[MemoryRecords](MemoryRecords)
                    .dump_json(self.__records, indent=2)
                    .decode("utf-8")
                )
        return "Remembered"

    @tool
    def recall(self):
        """Recall all the things you remembered"""
        try:
            with open(self.__memory_cache, "r") as f:
                return f.read()
        except FileNotFoundError:
            return "I don't remember anything"



import os
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional, Sequence, overload
import logging
import uuid
from agentia.history import History
from agentia.llm import LLM, LLMOptions, LLMOptionsUnion
from agentia.llm.completion import ChatCompletion, Listeners
from agentia.llm.stream import ChatCompletionStream, ChatCompletionEvents
from agentia.spec.chat import AssistantMessage, ToolMessage
from agentia.tools.tools import Tool, ToolSet
from agentia.spec import NonSystemMessage, UserMessage, MessagePartText, ObjectType
from dataclasses import asdict
from agentia.tools.mcp import MCPContext

if TYPE_CHECKING:
    from agentia.plugins.skills import Skills

type Events = Literal["finish"]


class Agent:
    def __init__(
        self,
        model: str | None = None,
        tools: Sequence[Tool] | None = None,
        id: str | None = None,
        instructions: str | None = None,
        options: LLMOptionsUnion | None = None,
        skills: Sequence[Path | str] | "Skills" | bool = False,
    ) -> None:
        from agentia.plugins.skills import Skills

        self.id = str(id or uuid.uuid4())
        self.options = (
            LLMOptions(**options)
            if isinstance(options, dict)
            else (options or LLMOptions())
        )
        if skills is not False:
            if not tools:
                tools = []
            if any(isinstance(t, Skills) for t in tools):
                raise ValueError(
                    "Cannot add a Skills plugin when another Skills plugin is already added"
                )
            if skills is True:
                tools = list(tools) + [Skills()]
            elif isinstance(skills, Skills):
                tools = list(tools) + [skills]
            elif isinstance(skills, Sequence):
                tools = list(tools) + [Skills(search_paths=skills)]
        if tools:
            self.options.tools = ToolSet(tools)
        if not model:
            model = os.getenv("AGENTIA_DEFAULT_MODEL", "openai/gpt-5-mini")
        self.llm = LLM(model)
        self.llm._agent = self
        self.history = History()
        if instructions:
            self.history.add_instructions(instructions)
        if self.options.tools and isinstance(self.options.tools, ToolSet):
            self.history.add_instructions(self.options.tools.get_instructions())
        self.log = logging.getLogger(f"agentia.agent")
        self._mcp_context: Optional[MCPContext] = None
        self.__on_finish = Listeners()

    def __add_prompt(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
    ) -> None:
        if isinstance(prompt, str):
            self.history.add(UserMessage(content=[MessagePartText(text=prompt)]))
        elif not isinstance(prompt, (list, Sequence)):
            self.history.add(prompt)
        else:
            self.history.add(*prompt)

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[False] = False,
        events: Literal[False] = False,
        options: LLMOptionsUnion | None = None,
    ) -> ChatCompletion: ...

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[True],
        events: Literal[False] = False,
        options: LLMOptionsUnion | None = None,
    ) -> ChatCompletionStream: ...

    @overload
    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: Literal[True],
        events: Literal[True],
        options: LLMOptionsUnion | None = None,
    ) -> ChatCompletionEvents: ...

    def run(
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        /,
        stream: bool = False,
        events: bool = False,
        options: LLMOptionsUnion | None = None,
    ) -> ChatCompletion | ChatCompletionStream | ChatCompletionEvents:
        self.__add_prompt(prompt)
        options_merged = LLMOptions()
        for k, v in asdict(self.options).items():
            setattr(options_merged, k, v)
        if options:
            options_dict = (
                asdict(options) if isinstance(options, LLMOptions) else options
            )
            for k, v in options_dict.items():
                setattr(options_merged, k, v)
        if stream:
            x = self.llm.stream(
                self.history.get(), events=events, options=options_merged
            )
        else:
            assert not events, "events=True is only supported with stream=True"
            x = self.llm.generate(self.history.get(), options=options_merged)

        def on_finish():
            for m in x.new_messages:
                assert isinstance(m, (AssistantMessage, ToolMessage, UserMessage))
                self.history.add(m)

        x.on_finish.on(on_finish)

        return x

    async def generate_object[T: ObjectType](
        self,
        prompt: str | NonSystemMessage | Sequence[NonSystemMessage],
        type: type[T],
    ) -> T:
        self.__add_prompt(prompt)
        msgs = await self.llm._generate_object_impl(
            self.history.get(), type, options=self.options
        )

        for m in msgs:
            assert isinstance(m, (AssistantMessage, ToolMessage, UserMessage))
            self.history.add(m)

        assert isinstance(msgs[-1], AssistantMessage)

        return msgs[-1].parse(type)

    async def __aenter__(self):
        self._mcp_context = MCPContext()
        await self._mcp_context.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        assert self._mcp_context is not None
        await self._mcp_context.__aexit__(exc_type, exc_val, exc_tb)
        self._mcp_context = None

    def emit(self, event: Events, *args, **kwargs):
        match event:
            case "finish":
                self.__on_finish.emit()

    def on(self, event: Events, listener=None):
        if not listener:

            def decorator(func):
                self.on(event, func)
                return func

            return decorator
        match event:
            case "finish":
                self.__on_finish.on(listener)

    def off(self, event: Events, listener=None):
        if not listener:

            def decorator(func):
                self.off(event, func)
                return func

            return decorator
        match event:
            case "finish":
                self.__on_finish.off(listener)



from . import Plugin, tool
from pathlib import Path
from typing import Any, Sequence, override
import frontmatter
from pydantic import BaseModel, Field
import subprocess
import os

INSTRUCTIONS_TEMPLATE = """\
You have access to a set of skills, each with its own instructions, capabilities, resources, and scripts for completing tasks within its spcific domain.

# AVAILABLE SKILLS
{skills}

# SKILL USAGE GUIDELINES
When a task falls within the domain of a skill,
1. Use `load_skill` to get the skill instructions and other detailed information.
2. Follow the skill instructions carefully to complete the task.
3. You can also use the resources and scripts provided by the skill to help you complete the task.

IMPORTANT: Load only the necessary skills to complete the task.
"""


class Skill(BaseModel):
    path: Path
    name: str
    description: str
    instructions: str
    resource_paths: list[str] = Field(default_factory=list)
    script_paths: list[str] = Field(default_factory=list)
    loaded: bool = False

    def to_dict(self) -> dict[str, Any]:
        resource_paths: list[str | dict[str, Any]] = []
        for r in self.resource_paths:
            if r.lower().endswith((".md", ".markdown")):
                content = (self.path / r).read_text(encoding="utf-8")
                post = frontmatter.loads(content)
                resource_paths.append({"path": r, **post.metadata})
            else:
                resource_paths.append(r)
        doc = self.model_dump()
        del doc["path"]
        doc["resource_paths"] = resource_paths
        return doc

    def load_resource(self, rel_path: str) -> str:
        if rel_path not in self.resource_paths:
            raise ValueError(f"Resource '{rel_path}' not found in skill '{self.name}'")
        resource_path = self.path / rel_path
        return resource_path.read_text(encoding="utf-8")

    def run_script(self, rel_path: str, args: list[str]) -> tuple[int, str, str]:
        if rel_path not in self.script_paths:
            raise ValueError(f"Script '{rel_path}' not found in skill '{self.name}'")
        script_path = self.path / rel_path
        cmd = ["python", str(script_path), *args]
        # if script has a sheband and is executable, run it directly
        if script_path.read_text(encoding="utf-8").startswith("#!"):
            if os.access(script_path, os.X_OK):
                cmd.pop(0)
        # run the script with the given arguments and return the output and stderr
        result = subprocess.run(cmd, capture_output=True, text=True, cwd=os.getcwd())
        return result.returncode, result.stdout, result.stderr


class Skills(Plugin):
    def __init__(self, search_paths: Sequence[Path | str] | None = None) -> None:
        if not search_paths:
            # default search path
            search_paths = [
                Path.cwd() / ".skills",
                Path.cwd() / ".agentia" / "skills",
                Path.home() / ".local" / "share" / "agentia" / "skills",
            ]
        self.__search_paths = [Path(p) for p in search_paths]
        # Load skills metadata from search paths
        self.skills: dict[str, Skill] = {}
        for path in self.__search_paths:
            if not path.is_dir():
                continue
            for skill in self.__load_skill_recursive(path, no_ignore=True):
                # assert skill.name not in self.skills, f"Duplicate skill name '{skill.name}' found in '{skill.path}'"
                self.skills[skill.name] = skill

    def add_skill(self, skill: Skill) -> None:
        if skill.name in self.skills:
            raise ValueError(f"A skill with the name '{skill.name}' already exists")
        self.skills[skill.name] = skill

    def __load_skill_recursive(self, path: Path, no_ignore=False) -> list[Skill]:
        assert path.is_dir(), f"'{path}' is not a directory"
        if not no_ignore and (path.name.startswith(".") or path.name.startswith("_")):
            return []
        if (path / "SKILL.md").is_file():
            return [self.__load_skill(path)]
        skills = []
        for child in path.iterdir():
            if child.is_dir():
                skills.extend(self.__load_skill_recursive(child))
        return skills

    def __load_skill(self, path: Path) -> Skill:
        path = path.resolve()
        md = path / "SKILL.md"
        assert md.is_file(), f"Skill metadata file '{md}' does not exist"
        doc = frontmatter.loads(md.read_text(encoding="utf-8"))
        name = doc.metadata.get("name", path.stem)
        assert isinstance(name, str), f"Skill name must be a string in '{md}'"
        description = doc.metadata.get("description", "")
        assert isinstance(
            description, str
        ), f"Skill description must be a string in '{md}'"
        instructions = doc.content
        # resources
        resources: list[str] = []
        supported_extensions = [".md", ".json", ".yaml", ".yml", ".csv", ".xml", ".txt"]
        excluded_files = {"SKILL.md"}
        for ext in supported_extensions:
            for p in (path / "resources").rglob(f"*{ext}"):
                if p.is_file() and p.name not in excluded_files:
                    try:
                        p = p.resolve()
                        rel_path = p.relative_to(path)
                        resources.append(rel_path.as_posix())
                    except Exception as e:
                        ...
        # scripts
        scripts: list[str] = []
        for p in (path / "scripts").rglob(f"*.py"):
            if p.is_file() and p.name != "__init__.py":
                try:
                    p = p.resolve()
                    rel_path = p.relative_to(path)
                    scripts.append(rel_path.as_posix())
                except Exception as e:
                    ...
        return Skill(
            path=path,
            name=name,
            description=description,
            instructions=instructions,
            resource_paths=resources,
            script_paths=scripts,
        )

    @override
    def get_instructions(self) -> str:
        s = ""
        for skill in self.skills.values():
            s += f"- **{skill.name}**:\n{skill.description}\n\n"
        return INSTRUCTIONS_TEMPLATE.format(skills=s)

    @tool(name="list_skills")
    def list_skills(self) -> Any:
        """
        Get the list of available skills, their high-level descriptions, and whether they are loaded.
        Use this tool to discover what skills are available and their descriptions.
        Only use this tool when you need to refresh your memory.
        """
        skills = list(self.skills.values())
        skill_docs = []
        for skill in skills:
            skill_docs.append(
                {
                    "name": skill.name,
                    "description": skill.description,
                    "loaded": skill.loaded,
                }
            )
        return skill_docs

    @tool(name="load_skill")
    def load_skill(self, skill_name: str) -> Any:
        """
        Load the instructions and capabilities of a skill by its name.

        Also lists the resources and scripts provided by the skill, if any, so that you can use them later.

        Call this tool (only once) when you need to perform a task within the skill's domain.
        """
        if skill_name not in self.skills:
            return {"error": f"Skill '{skill_name}' not found"}
        skill = self.skills[skill_name]
        return skill.to_dict()

    @tool(name="load_skill_resource")
    def load_skill_resource(self, skill_name: str, resource_path: str) -> Any:
        """
        Load a resource provided by a skill.

        `resource_path` is the path of the resource within the skill directory.

        Call this tool when you need to use a specific resource provided by a skill.
        """
        if skill_name not in self.skills:
            return {"error": f"Skill '{skill_name}' not found"}
        skill = self.skills[skill_name]
        try:
            content = skill.load_resource(resource_path)
            return {"content": content}
        except Exception as e:
            return {"error": str(e)}

    @tool(name="run_skill_script")
    def run_skill_script(
        self, skill_name: str, script_path: str, args: list[str]
    ) -> Any:
        """
        Run a script provided by a skill with the given arguments.

        `script_path` is the path of the script within the skill's `scripts` directory.

        Call this tool when you need to execute a specific script provided by a skill to complete a task.
        """
        if skill_name not in self.skills:
            return {"error": f"Skill '{skill_name}' not found"}
        skill = self.skills[skill_name]
        try:
            code, stdout, stderr = skill.run_script(script_path, args)
            if code != 0:
                return {
                    "error": f"Script exited with code {code}",
                    "stdout": stdout,
                    "stderr": stderr,
                }
            else:
                doc = {"returncode": code, "stdout": stdout}
                if stderr:
                    doc["stderr"] = stderr
                return doc
        except Exception as e:
            return {"error": str(e)}



from pathlib import Path
from agentia.utils.skill_learner import SkillInfo
from . import Plugin, tool
from typing import Annotated


class SkillLearner(Plugin):
    def __init__(self, save_dir: Path | str | None = None) -> None:
        if save_dir is None:
            save_dir = Path.cwd() / ".skills"
        self.save_dir = Path(save_dir)

    @tool(name="generate_skill")
    async def generate_skill(
        self,
        name: Annotated[
            str,
            "The name of the skill. lowercase, no space, hyphen separated. Must be unique among all skills.",
        ],
        skill: SkillInfo,
    ):
        """
        Create/generate a skill based on your prior knowledge. The skills can then be reused either by you or other LLMs to solve problems or offer guidance within the scope of the skill.
        """
        from agentia.plugins.skills import Skills

        assert self.llm and self.llm._active_tools
        existing_skills: set[str] = set()
        skills: Skills | None = self.llm._active_tools.get_plugin(Skills)
        if skills:
            existing_skills = set(skills.skills.keys())

        s = skill.save(name=name, out=self.save_dir / name)

        if name in existing_skills:
            raise ValueError(f"A skill with the name '{name}' already exists")

        if skills:
            skills.add_skill(s)

        return {"name": s.name, "description": s.description, "loaded": False}

    @tool(name="learn_skill_from_documents")
    async def learn_skill_from_documents(
        self,
        name: Annotated[
            str,
            "The name of the skill. lowercase, no space, hyphen separated. Must be unique among all skills.",
        ],
        docs: Annotated[
            list[str],
            "A list of documents to learn the skill from. Each document can be a pdf/txt/markdown file path or a URL.",
        ],
        prompt: Annotated[
            str,
            "Additional instructions or information to guide the learning of the skill. e.g. what skill to learn or focus on, etc.",
        ],
    ):
        """Learn or acquire a skill from a list of documents"""
        from agentia.utils.skill_learner import learn_skill_from_documents
        from agentia.plugins.skills import Skills

        assert self.llm and self.llm._active_tools

        for d in docs:
            if not d.startswith(("http://", "https://")):
                if not Path(d).is_file():
                    raise ValueError(f"Document '{d}' does not exist")

        existing_skills: set[str] = set()
        skills: Skills | None = self.llm._active_tools.get_plugin(Skills)
        if skills:
            existing_skills = set(skills.skills.keys())

        if name in existing_skills:
            raise ValueError(f"A skill with the name '{name}' already exists")

        path = self.save_dir / name
        path.mkdir(parents=True, exist_ok=True)

        skill = await learn_skill_from_documents(
            docs=docs, name=name, prompt=prompt, out=path, model=self.llm.model
        )

        if skills:
            skills.add_skill(skill)

        return {"name": skill.name, "description": skill.description, "loaded": False}



from datetime import datetime
from typing import Annotated, Literal, Sequence
from pydantic import AliasChoices, BaseModel, Field
from .base import *


class StreamPartTextStart(BaseModel):
    type: Literal["text-start"] = "text-start"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartTextDelta(BaseModel):
    type: Literal["text-delta"] = "text-delta"
    delta: str
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartTextEnd(BaseModel):
    type: Literal["text-end"] = "text-end"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningStart(BaseModel):
    type: Literal["reasoning-start"] = "reasoning-start"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningDelta(BaseModel):
    type: Literal["reasoning-delta"] = "reasoning-delta"
    delta: str
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningEnd(BaseModel):
    type: Literal["reasoning-end"] = "reasoning-end"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartStreamStart(BaseModel):
    type: Literal["stream-start"] = "stream-start"
    id: str | None = None
    timestamp: datetime | None = None
    model_id: str | None = Field(
        default=None,
        validation_alias=AliasChoices("modelId", "model_id"),
        serialization_alias="modelId",
    )


class StreamPartStreamEnd(BaseModel):
    type: Literal["finish"] = "finish"
    usage: Usage
    finish_reason: FinishReason = Field(
        validation_alias=AliasChoices("finishReason", "finish_reason"),
        serialization_alias="finishReason",
    )


class Annotations(BaseModel):
    type: Literal["annotations"] = "annotations"
    annotations: Sequence[Annotation]

    def __init__(
        self,
        annotations: Sequence[Annotation],
        *,
        type: Literal["annotations"] = "annotations",
    ):
        super().__init__(type=type, annotations=annotations)


type StreamPart = Annotated[
    StreamPartTextStart
    | StreamPartTextDelta
    | StreamPartTextEnd
    | StreamPartReasoningStart
    | StreamPartReasoningDelta
    | StreamPartReasoningEnd
    | ToolCall
    | ToolResult
    | Annotation
    | StreamPartStreamStart
    | StreamPartStreamEnd,
    Field(discriminator="type"),
]

__all__ = [
    "StreamPartTextStart",
    "StreamPartTextDelta",
    "StreamPartTextEnd",
    "StreamPartReasoningStart",
    "StreamPartReasoningDelta",
    "StreamPartReasoningEnd",
    "StreamPartStreamStart",
    "StreamPartStreamEnd",
    "StreamPart",
]



import abc
from typing import Any, Optional, Self, TYPE_CHECKING

if TYPE_CHECKING:
    from agentia.llm import LLM
    from agentia.agent import Agent


class PluginInitError(RuntimeError):
    def __init__(self, plugin: str, original: Exception) -> None:
        self.plugin = plugin
        self.msg = str(original)
        self.original = original
        super().__init__(f"Plugin {plugin} failed to initialize: {self.msg}")


class Plugin(abc.ABC):
    NAME: str | None = None
    _BUILTIN_ID: str | None = None

    def get_instructions(self) -> str | None:
        """Get the instructions for using this plugin. This will be included in the agent's system prompt."""
        return None

    @classmethod
    def name(cls) -> str:
        if cls.NAME:
            return cls.NAME
        name = cls.__name__
        if name.endswith("Plugin"):
            name = name[:-6]
        return name

    @classmethod
    def id(cls) -> str:
        if cls._BUILTIN_ID:
            return cls._BUILTIN_ID
        return cls.name().lower()

    def __init__(self, *args: Any, **kwargs: Any):
        self.config: dict[str, Any] = {}
        self.llm: Optional["LLM"] = None
        self.agent: Optional["Agent"] = None

    @classmethod
    def instantiate(cls, config: dict[str, Any]) -> Self:
        """Instantiate the plugin with the given config."""
        try:
            obj = cls(**(config or {}))
            obj.config = config
            return obj
        except Exception as e:
            raise PluginInitError(cls.id(), e) from e

    async def init(self):
        """
        Initialize the plugin before running the agent.
        This may involve creating API clients, login, and verify auth tokens.

        For OAuth:
            Login steps here can only access to CLI inputs.
            You may also want to override `__options__` so that user can login on the dashboard site.
        """
        pass

    @classmethod
    def validate_config(cls, config: dict[str, Any]): ...



from __future__ import annotations

import logging
import os

LOGGER = logging.getLogger("agentia")


def init_logging(level: logging._Level = logging.INFO):
    logging.basicConfig(
        level=level, format="[%(asctime)s][%(levelname)-8s][%(name)s] %(message)s"
    )
    logging.getLogger("httpx").setLevel(logging.WARNING)


from dotenv import load_dotenv

from . import utils

from .agent import Agent
from .spec import *
from .utils.decorators import magic, ImageUrl
from .utils.decorators import tool
from .plugins import Plugin, PluginInitError
from .plugins.skills import Skills, Skill
from . import llm, plugins
from .tools import ProviderTool, Tool, Tools, MCP, MCPContext
from .history import History
from .llm import LLMOptions
from .llm.completion import ChatCompletion
from .llm.stream import (
    TextStream,
    ReasoningStream,
    MessageStream,
    ChatCompletionStream,
    ChatCompletionEvents,
)

if os.environ.get("AGENTIA_PATCH", "true").lower() in ("true", "1", "yes"):
    from .utils.patches import patch_all

    patch_all()

__all__ = [
    # submodules
    "plugins",
    "utils",
    "llm",
    # agent
    "Agent",
    "History",
    # llm
    "ChatCompletion",
    "TextStream",
    "ReasoningStream",
    "MessageStream",
    "ChatCompletionStream",
    "ChatCompletionEvents",
    "LLMOptions",
    # spec
    "ProviderOptions",
    "ProviderMetadata",
    "DataContent",
    "FinishReason",
    "ToolChoice",
    "Usage",
    "FunctionTool",
    "ProviderDefinedTool",
    "Tool",
    "ToolCall",
    "ToolResult",
    "File",
    "Annotation",
    "MessageBase",
    "SystemMessage",
    "MessagePartBase",
    "MessagePartText",
    "MessagePartReasoning",
    "MessagePartFile",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "MessagePart",
    "UserMessagePart",
    "AssistantMessagePart",
    "UserMessage",
    "ObjectType",
    "AssistantMessage",
    "ToolMessage",
    "Message",
    "NonSystemMessage",
    "Prompt",
    "ResponseFormatText",
    "ResponseFormatJson",
    "ResponseFormat",
    "StreamPartTextStart",
    "StreamPartTextDelta",
    "StreamPartTextEnd",
    "StreamPartReasoningStart",
    "StreamPartReasoningDelta",
    "StreamPartReasoningEnd",
    "StreamPartStreamStart",
    "StreamPartStreamEnd",
    "StreamPart",
    # plugins
    "Plugin",
    "PluginInitError",
    "ToolResult",
    "Skills",
    "Skill",
    # utils
    "init_logging",
    "magic",
    "tool",
    "ImageUrl",
    "ProviderTool",
    "Tool",
    "Tools",
    "MCP",
    "MCPContext",
    "load_dotenv",
]



from .mcp import MCP, MCPContext
from .tools import Tool, Tools, ProviderTool, FileResult
from .plugin import Plugin, PluginInitError
from ..utils.decorators import tool

__all__ = [
    "MCP",
    "MCPContext",
    "Tool",
    "Tools",
    "ProviderTool",
    "Plugin",
    "PluginInitError",
    "FileResult",
    "tool",
]



from typing import TYPE_CHECKING, Annotated, Any, Literal
from pydantic import AliasChoices, BaseModel, Field, JsonValue, HttpUrl


type ProviderOptions = dict[str, JsonValue]
type ProviderMetadata = dict[str, dict[str, JsonValue]]


type DataContent = bytes | str | HttpUrl

type FinishReason = Literal[
    "stop", "length", "content-filter", "tool-calls", "error", "other", "unknown"
]


type ToolChoice = Literal["auto", "none", "required"] | str


class Usage(BaseModel):
    input_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("inputTokens", "input_tokens"),
        serialization_alias="inputTokens",
    )
    output_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("outputTokens", "output_tokens"),
        serialization_alias="outputTokens",
    )
    total_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("totalTokens", "total_tokens"),
        serialization_alias="totalTokens",
    )
    reasoning_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("reasoningTokens", "reasoning_tokens"),
        serialization_alias="reasoningTokens",
    )
    cached_input_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("cachedInputTokens", "cached_input_tokens"),
        serialization_alias="cachedInputTokens",
    )

    def __add__(self, y: "Usage") -> "Usage":
        x = Usage(**self.model_dump())
        if y.input_tokens:
            x.input_tokens = (x.input_tokens or 0) + y.input_tokens
        if y.output_tokens:
            x.output_tokens = (x.output_tokens or 0) + y.output_tokens
        if y.total_tokens:
            x.total_tokens = (x.total_tokens or 0) + y.total_tokens
        if y.reasoning_tokens:
            x.reasoning_tokens = (x.reasoning_tokens or 0) + y.reasoning_tokens
        if y.cached_input_tokens:
            x.cached_input_tokens = (x.cached_input_tokens or 0) + y.cached_input_tokens
        return x


class FunctionTool(BaseModel):
    type: Literal["function"] = "function"
    name: str
    input_schema: JsonValue = Field(
        validation_alias=AliasChoices("inputSchema", "input_schema"),
        serialization_alias="inputSchema",
    )
    description: str | None = None
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )

    def to_openai_schema(self) -> Any:
        return {
            "type": self.type,
            "name": self.name,
            "description": self.description,
            "parameters": self.input_schema,
            "strict": True,
        }


class ProviderDefinedTool(BaseModel):
    type: Literal["provider-defined"] = "provider-defined"
    id: str
    name: str
    args: dict[str, JsonValue] = Field(default_factory=dict)

    def to_openai_schema(self) -> Any:
        return {"type": self.id, **self.args}


type Tool = Annotated[FunctionTool | ProviderDefinedTool, Field(discriminator="type")]


class ToolCall(BaseModel):
    type: Literal["tool-call"] = "tool-call"

    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    """The identifier of the tool call. It must be unique across all tool calls."""

    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    """The name of the tool that should be called."""

    input: JsonValue
    """JSON object with the tool call arguments. Must match the parameters schema of the tool."""

    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    """
    Whether the tool call will be executed by the provider.
    If this flag is not set or is false, the tool call will be executed by the client.
    """

    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )
    """Additional provider-specific metadata for the tool call."""


class ToolResult(BaseModel):
    type: Literal["tool-result"] = "tool-result"

    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    """The ID of the tool call that this result is associated with."""

    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    """Name of the tool that generated this result."""

    result: JsonValue
    """Result of the tool call. This is a JSON-serializable object."""

    is_error: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("isError", "is_error"),
        serialization_alias="isError",
    )
    """Optional flag if the result is an error or an error message."""

    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    """
    Whether the tool result was generated by the provider.
    If this flag is set to true, the tool result was generated by the provider.
    If this flag is not set or is false, the tool result was generated by the client.
    """

    preliminary: bool | None = None
    """
    Whether the tool result is preliminary.

    Preliminary tool results replace each other, e.g. image previews.
    There always has to be a final, non-preliminary tool result.

    If this flag is set to true, the tool result is preliminary.
    If this flag is not set or is false, the tool result is not preliminary.
    """

    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )
    """Additional provider-specific metadata for the tool result."""


class File(BaseModel):
    type: Literal["file"] = "file"

    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )
    """
    The IANA media type of the file, e.g. `image/png` or `audio/mp3`.

    @see https://www.iana.org/assignments/media-types/media-types.xhtml
    """

    data: DataContent
    """
    Generated file data as base64 encoded strings or binary data.

    The file data should be returned without any unnecessary conversion.
    If the API returns base64 encoded strings, the file data should be returned
    as base64 encoded strings. If the API returns binary data, the file data should
    be returned as binary data.
    """

    def to_url(self) -> str:
        if isinstance(self.data, bytes):
            import base64

            return (
                f"data:{self.media_type};base64,{base64.b64encode(self.data).decode()}"
            )
        elif isinstance(self.data, str):
            return self.data
        else:
            return str(self.data)


class Annotation(BaseModel):
    type: Literal["annotation"] = "annotation"

    end: int
    """The index of the last character of the URL citation in the message."""

    start: int
    """The index of the first character of the URL citation in the message."""

    title: str
    """The title of the web resource."""

    url: str
    """The URL of the web resource."""


__all__ = [
    "ProviderOptions",
    "ProviderMetadata",
    "DataContent",
    "FinishReason",
    "ToolChoice",
    "Usage",
    "FunctionTool",
    "ProviderDefinedTool",
    "Tool",
    "ToolCall",
    "ToolResult",
    "File",
    "Annotation",
]



from contextlib import redirect_stderr
from . import Plugin, tool
from typing import Annotated
import traceback


class CodeRunner(Plugin):
    @tool
    def execute(self, python_code: Annotated[str, "The python code to run."]):
        """Execute python code and return the result. The expression must be an valid python expression that can be execuated by `eval()`."""
        from contextlib import redirect_stdout
        import io

        out = io.StringIO()
        err = io.StringIO()
        with redirect_stdout(out):
            with redirect_stderr(err):
                try:
                    exec(python_code, globals())
                    o = out.getvalue()
                    e = err.getvalue()
                    result = {
                        "stdout": o,
                        "stderr": e,
                        "success": True,
                    }
                except Exception as ex:
                    o = out.getvalue()
                    e = err.getvalue()
                    result = {
                        "stdout": o,
                        "stderr": e,
                        "success": False,
                        "error": str(ex),
                        "traceback": repr(traceback.format_exc()),
                    }

        return result



import asyncio
from typing import Any, AsyncGenerator, Generator, Optional, TYPE_CHECKING
import os
from agentia.spec.base import FinishReason, Usage
from agentia.spec.chat import AssistantMessage, Message, ToolMessage

if TYPE_CHECKING:
    from agentia.agent import Agent


class Listeners:
    def __init__(self):
        self.__listeners = []

    def on(self, listener):
        self.__listeners.append(listener)

    def off(self, listener):
        self.__listeners.remove(listener)

    def emit(self, *args, **kwargs):
        for listener in self.__listeners:
            listener(*args, **kwargs)


class ChatCompletion:
    def __init__(
        self,
        gen: AsyncGenerator[AssistantMessage | ToolMessage, None],
        agent: Optional["Agent"],
    ):
        async def __gen() -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
            async for msg in gen:
                yield msg
            self._on_finish()

        self.__gen = __gen()
        self.usage = Usage()
        self.finish_reason: FinishReason | None = None
        self.new_messages: list[Message] = []
        self.on_finish = Listeners()
        self.on_new_message = Listeners()
        self.agent = agent

    def _on_finish(self):
        self.on_finish.emit()
        if self.agent:
            self.agent.emit("finish")

    def add_new_message(self, msg: Message):
        self.new_messages.append(msg)
        self.on_new_message.emit(msg)

    def __aiter__(self) -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
        return self.__gen

    async def __wait_for_completion(self) -> AssistantMessage:
        async for msg in self.__gen:
            ...
        m = self.new_messages[-1]
        assert isinstance(m, AssistantMessage)
        return m

    def __await__(self):
        return self.__wait_for_completion().__await__()

    def __iter__(self):
        return async_gen_to_sync(self.__gen)


def async_gen_to_sync[T](agen: AsyncGenerator[T, None]) -> Generator[T, None, None]:
    try:
        loop = asyncio.get_running_loop()
        close = False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio._set_running_loop(loop)
        close = True
    try:
        while True:
            try:
                item = loop.run_until_complete(agen.__anext__())
                yield item
            except StopAsyncIteration:
                break
    finally:
        if close:
            loop.close()
            asyncio._set_running_loop(None)



import asyncio
import inspect
import os
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

from pydantic import AliasChoices, BaseModel, Field, JsonValue, ValidationError
from agentia.tools.mcp import MCP
import agentia.spec as spec
from agentia.utils.decorators import ToolFuncParam
from inspect import Parameter
from uuid import uuid4
from pydantic.type_adapter import TypeAdapter

NAME_TAG = "agentia_tool_name"
DISPLAY_NAME_TAG = "agentia_tool_display_name"
IS_TOOL_TAG = "agentia_tool_is_tool"
DESCRIPTION_TAG = "agentia_tool_description"
METADATA_TAG = "agentia_tool_metadata"

if TYPE_CHECKING:
    from ..llm import LLM
    from agentia.plugins import Plugin
    from agentia.agent import Agent


class _BaseTool:
    def __init__(
        self,
        name: str,
        schema: Any,
        display_name: str | None = None,
        description: str | None = None,
        metadata: Any | None = None,
    ):
        self.name = name
        self.display_name = display_name
        self.description = description
        self.schema = schema
        self.metadata = metadata


class _PythonFunctionTool(_BaseTool):
    def __init__(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        if plugin:
            name = getattr(f, NAME_TAG, None) or f"{plugin.name()}__{f.__name__}"
            display_name: str = (
                getattr(f, DISPLAY_NAME_TAG, None) or f"{plugin.name()}@{f.__name__}"
            )
        else:
            name = getattr(f, NAME_TAG, f.__name__)
            display_name = getattr(f, DISPLAY_NAME_TAG, name)
        description: str = getattr(f, DESCRIPTION_TAG, f.__doc__) or ""
        metadata = getattr(f, METADATA_TAG, None)
        params = [
            ToolFuncParam(p, f.__name__)
            for p in inspect.signature(f).parameters.values()
            if p.name != "self"
        ]
        schema = _PythonFunctionTool.get_json_schema(
            name=name, description=description, fparams=params
        )
        super().__init__(
            name=name,
            schema=schema,
            display_name=display_name,
            description=description,
            metadata=metadata,
        )
        self.func = f
        self.plugin = plugin
        self.params = params

    @staticmethod
    def get_json_schema(
        name: str, description: str | None, fparams: list[ToolFuncParam]
    ) -> Any:
        params: Any = {"type": "object", "properties": {}, "required": []}
        for p in fparams:
            if p.required:
                params["required"].append(p.name)
            params["properties"][p.name] = p.schema
        return {
            "name": name,
            "description": description or "",
            "parameters": params,
        }

    def process_json_args(self, llm: "LLM", args: Any):
        from agentia.agent import Agent
        from ..llm import LLM

        p_args: list[Any] = []
        kw_args: dict[str, Any] = {}

        def get_value(p: ToolFuncParam, name: str, default: Any, is_model: bool):
            if name not in args:
                return default
            if is_model:
                return p.type(**args[name])
            else:
                v = args[name]
                try:
                    return TypeAdapter(p.type).validate_python(v)
                except ValidationError as e:
                    raise ValueError(
                        f"Error parsing argument '{name}': {e}. Please retry with a valid value."
                    ) from e

        assert self.params is not None
        for p in self.params:
            is_model = inspect.isclass(p.type) and issubclass(p.type, BaseModel)
            match p.param.kind:
                case Parameter.POSITIONAL_ONLY if p.param.annotation == Agent:
                    if llm._agent is not None:
                        p_args.append(llm._agent)
                case Parameter.POSITIONAL_ONLY if p.param.annotation == LLM:
                    p_args.append(llm)
                case Parameter.POSITIONAL_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    p_args.append(get_value(p, p.name, default, is_model))
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY if (
                    p.param.annotation == Agent
                ):
                    if llm._agent is not None:
                        kw_args[p.name] = llm._agent
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    kw_args[p.name] = get_value(p, p.name, default, is_model)
                case other:
                    raise ValueError(f"{other} is not supported")
        return p_args, kw_args


class _MCPTool(_BaseTool):
    def __init__(self, name: str, schema: Any, server: MCP):
        super().__init__(name=name, schema=schema)
        self.server = server


class ProviderTool(BaseModel):
    name: str
    args: dict[str, JsonValue] | None = None


type Tool = Plugin | Callable[..., Any] | ProviderTool | MCP
type Tools = Sequence[Tool]


class FileResult(BaseModel):
    id: str | None = None
    data: spec.DataContent
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )


class ToolSet:
    def __init__(self, tools: Tools):
        from agentia.plugins import Plugin

        self.all_tools = tools
        self.plugins: dict[str, Plugin] = {}
        self.mcp_servers: dict[str, MCP] = {}
        self.provider_tools: dict[str, ProviderTool] = {}
        self.__tools: dict[str, _BaseTool] = {}
        for t in tools:
            if isinstance(t, Plugin):
                self.__add_plugin(t)
            elif isinstance(t, MCP):
                self.__add_mcp_server(t)
            elif isinstance(t, ProviderTool):
                self.provider_tools[t.name] = t
            else:
                assert inspect.isfunction(
                    t
                ), "Expected a function, MCP server object, Plugin, or ProviderTool"
                self.__add_function(t)
        self.__initialized = False

    async def init(self, llm: "LLM", agent: Optional["Agent"] = None):
        if self.__initialized:
            return
        self.__initialized = True
        for plugin in self.plugins.values():
            try:
                plugin.llm = llm
                plugin.agent = agent
                await plugin.init()
            except Exception as e:
                from . import PluginInitError

                raise PluginInitError(plugin.id(), e) from e
        for server in self.mcp_servers.values():
            await server.init(llm=llm, agent=agent)
            tools = server.get_tools()
            for tool in tools:
                self.__tools[tool.name] = tool

    def add(self, t: Union[Callable[..., Any], MCP, "Plugin"]) -> None:
        from agentia.plugins import Plugin

        if isinstance(t, Plugin):
            self.__add_plugin(t)
        elif isinstance(t, MCP):
            self.__add_mcp_server(t)
        else:
            assert inspect.isfunction(
                t
            ), "Expected a function, MCP server object, or Plugin"
            self.__add_function(t)

    def __add_function(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        t = _PythonFunctionTool(f=f, plugin=plugin)
        self.__tools[t.name] = t

    def __add_plugin(self, p: "Plugin"):
        # Add all functions from the plugin
        for _n, method in inspect.getmembers(p, predicate=inspect.ismethod):
            if not getattr(method, IS_TOOL_TAG, False):
                continue
            self.__add_function(method, plugin=p)
        # Add the plugin to the list of plugins
        self.plugins[p.id()] = p

    def __add_mcp_server(self, server: MCP):
        self.mcp_servers[server.name] = server

    def get_plugin[T: "Plugin"](self, type: type[T]) -> Optional[T]:
        for p in self.plugins.values():
            if isinstance(p, type):
                return p
        return None

    def get_instructions(self) -> str:
        instructions = []
        for plugin in self.plugins.values():
            ins = plugin.get_instructions()
            if ins:
                ins = ins.strip()
                if ins:
                    instructions.append(ins)
        return "\n\n".join(instructions)

    def is_empty(self) -> bool:
        return len(self.__tools) == 0

    def get_schema(self) -> Sequence[spec.ProviderDefinedTool | spec.FunctionTool]:
        functions = [
            spec.FunctionTool(
                name=v.name,
                description=v.description or "",
                input_schema=v.schema["parameters"],
            )
            for (k, v) in self.__tools.items()
        ]
        provider_tools = [
            spec.ProviderDefinedTool(id=t.name, name=t.name, args=t.args or {})
            for t in self.provider_tools.values()
        ]
        return functions + provider_tools

    async def __run_python_tool(
        self,
        id: str,
        tool: _PythonFunctionTool,
        llm: "LLM",
        args: Any,
    ) -> tuple[spec.ToolResult, FileResult | None]:
        p_args, kw_args = tool.process_json_args(llm, args)
        output = tool.func(*p_args, **kw_args)  # type: ignore
        if inspect.isawaitable(output):
            output = await output
        # if output is a FileResult, transform the result
        file = None
        tool_vision = os.getenv("AGENTIA_EXPERIMENTAL_TOOL_FILES", None)
        if tool_vision and tool_vision != "0" and tool_vision.lower() != "false":
            if isinstance(output, FileResult):
                file = output
                if not file.id:
                    file.id = str(uuid4())
                output: Any = {
                    "file_id": file.id,
                    "media_type": file.media_type,
                    "hint": "The tool outputed a file with the given file_id. The file is attached below.",
                }
                d = file.data
                if isinstance(d, str) and (d.startswith(("http://", "https://"))):
                    output["url"] = d
        else:
            if isinstance(output, FileResult):
                output = output.model_dump()
                if isinstance(output["data"], str) and output["data"].startswith(
                    "data:"
                ):
                    del output["data"]
                output["hint"] = "The tool outputed a file with the given file_id."

        tr = spec.ToolResult(tool_call_id=id, tool_name=tool.name, result=output)
        return tr, file

    async def __run_mcp_tool(
        self, id: str, tool: _MCPTool, args: Any
    ) -> spec.ToolResult:
        result = await tool.server.run(tool.name, args)
        tr = spec.ToolResult(tool_call_id=id, tool_name=tool.name, result=result)
        return tr

    async def __run_one(
        self, llm: "LLM", c: spec.ToolCall
    ) -> tuple[spec.ToolResult, FileResult | None]:
        tool = self.__tools.get(c.tool_name, None)
        if not tool:
            raise ValueError(f"Tool {c.tool_name} not found")
        try:
            if isinstance(tool, _PythonFunctionTool):
                result, file = await self.__run_python_tool(
                    id=c.tool_call_id,
                    tool=tool,
                    llm=llm,
                    args=c.input,
                )
                return result, file
            elif isinstance(tool, _MCPTool):
                result = await self.__run_mcp_tool(
                    id=c.tool_call_id,
                    tool=tool,
                    args=c.input,
                )
                return result, None
            else:
                raise ValueError(f"Tool {c.tool_name} is of unknown type")
        except Exception as e:
            output: JsonValue = {"error": str(e)}
            r = spec.ToolResult(
                tool_call_id=c.tool_call_id, tool_name=c.tool_name, result=output
            )
            return r, None

    async def run(
        self, llm: "LLM", tool_calls: list[spec.ToolCall], parallel: bool
    ) -> tuple[spec.ToolMessage, list[spec.ToolResult], list[FileResult]]:
        if parallel:
            results = await asyncio.gather(
                *[self.__run_one(llm, c) for c in tool_calls]
            )
        else:
            results = []
            for c in tool_calls:
                results.append(await self.__run_one(llm, c))
        parts: list[spec.MessagePartToolResult] = []
        tool_results: list[spec.ToolResult] = []
        files: list[FileResult] = []
        for r, f in results:
            tool_results.append(r)
            parts.append(
                spec.MessagePartToolResult(
                    tool_call_id=r.tool_call_id,
                    tool_name=r.tool_name,
                    output=r.result,
                )
            )
            if f:
                files.append(f)
        return spec.ToolMessage(content=parts), tool_results, files



from enum import Enum
import inspect
from typing import Annotated, Any, Literal, Sequence, Type
from .base import *
from openai.lib._parsing._completions import to_strict_json_schema  # type: ignore
from pydantic import AliasChoices, BaseModel, Field, JsonValue


class MessageBase(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


class SystemMessage(MessageBase):
    role: Literal["system"] = "system"
    content: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        content: str,
        *,
        role: Literal["system"] = "system",
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(role=role, content=content, provider_options=provider_options)


class MessagePartBase(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


class MessagePartText(MessagePartBase):
    type: Literal["text"] = "text"
    text: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        text: str,
        *,
        type: Literal["text"] = "text",
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(type=type, text=text, provider_options=provider_options)


class MessagePartReasoning(MessagePartBase):
    type: Literal["reasoning"] = "reasoning"
    text: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        text: str,
        *,
        type: Literal["reasoning"] = "reasoning",
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(type=type, text=text, provider_options=provider_options)


class MessagePartFile(MessagePartBase):
    type: Literal["file"] = "file"
    data: DataContent
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )
    filename: str | None = None
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )

    def to_url(self) -> str:
        return File(data=self.data, media_type=self.media_type).to_url()


class MessagePartToolCall(MessagePartBase):
    type: Literal["tool-call"] = "tool-call"
    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    input: JsonValue
    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


class MessagePartToolResult(MessagePartBase):
    type: Literal["tool-result"] = "tool-result"
    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    output: JsonValue
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


type MessagePart = Annotated[
    MessagePartText
    | MessagePartReasoning
    | MessagePartFile
    | MessagePartToolCall
    | MessagePartToolResult,
    Field(discriminator="type"),
]

type UserMessagePart = Annotated[
    MessagePartText | MessagePartFile, Field(discriminator="type")
]

type AssistantMessagePart = Annotated[
    MessagePartText | MessagePartReasoning | MessagePartToolCall,
    Field(discriminator="type"),
]


class UserMessage(MessageBase):
    role: Literal["user"] = "user"
    content: Sequence[UserMessagePart] | UserMessagePart | str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        content: str | Sequence[UserMessagePart] | UserMessagePart,
        *,
        role: Literal["user"] = "user",
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(role=role, content=content, provider_options=provider_options)

    @property
    def content_list(self) -> list[UserMessagePart]:
        if isinstance(self.content, str):
            return [MessagePartText(text=self.content)]
        if not isinstance(self.content, (Sequence, list)):
            return [self.content]
        return list(self.content)

    @property
    def text(self) -> str:
        text = ""
        for part in self.content_list:
            if isinstance(part, MessagePartText):
                text += part.text
        return text


type ObjectType = BaseModel | Enum | str | int | float | bool | None | list[
    "ObjectType"
] | dict[str, "ObjectType"] | tuple["ObjectType", ...]


class _Result[X](BaseModel):
    result: X = Field(..., description="The result", title="Result")


class AssistantMessage(MessageBase):
    role: Literal["assistant"] = "assistant"
    content: Sequence[AssistantMessagePart] | AssistantMessagePart | str
    annotations: list[Annotation] | None = None
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        content: str | Sequence[AssistantMessagePart] | AssistantMessagePart,
        *,
        role: Literal["assistant"] = "assistant",
        annotations: list[Annotation] | None = None,
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(role=role, content=content, provider_options=provider_options)

    @property
    def content_list(self) -> list[AssistantMessagePart]:
        if isinstance(self.content, str):
            return [MessagePartText(text=self.content)]
        if not isinstance(self.content, Sequence):
            return [self.content]
        return list(self.content)

    @property
    def reasoning(self) -> str | None:
        text = ""
        for part in self.content:
            if isinstance(part, MessagePartReasoning):
                text += part.text
        return text

    @property
    def text(self) -> str:
        text = ""
        for part in self.content_list:
            if isinstance(part, MessagePartText):
                text += part.text
        return text

    def parse[T: ObjectType](self, return_type: type[T]) -> T:
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result

        json_string = ""
        for part in self.content:
            if isinstance(part, MessagePartText):
                json_string += part.text

        result = response_format.model_validate_json(json_string)
        if isinstance(result, _Result):
            return result.result
        else:
            return result


class ToolMessage(MessageBase):
    role: Literal["tool"] = "tool"
    content: Sequence[MessagePartToolResult]
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    def __init__(
        self,
        content: Sequence[MessagePartToolResult],
        *,
        role: Literal["tool"] = "tool",
        provider_options: ProviderOptions | None = None,
    ):
        super().__init__(role=role, content=content, provider_options=provider_options)


type Message = Annotated[
    SystemMessage | UserMessage | AssistantMessage | ToolMessage,
    Field(discriminator="role"),
]

type NonSystemMessage = Annotated[
    UserMessage | AssistantMessage | ToolMessage, Field(discriminator="role")
]

type Prompt = Sequence[Message]


class ResponseFormatText(BaseModel):
    type: Literal["text"] = "text"


class ResponseFormatJson(BaseModel):
    type: Literal["json"] = "json"
    json_schema: JsonValue | None = Field(
        default=None,
        validation_alias=AliasChoices("schema", "json_schema"),
        serialization_alias="schema",
    )
    name: str | None = None
    description: str | None = None

    @staticmethod
    def from_model[T: ObjectType](return_type: Type[T]) -> "ResponseFormatJson":
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result
        return ResponseFormatJson(
            json_schema=to_strict_json_schema(response_format),
            name=return_type.__name__,
            description=f"JSON object matching the schema of {return_type.__name__}",
        )


type ResponseFormat = Annotated[
    ResponseFormatText | ResponseFormatJson, Field(discriminator="type")
]


__all__ = [
    "MessageBase",
    "SystemMessage",
    "MessagePartBase",
    "MessagePartText",
    "MessagePartReasoning",
    "MessagePartFile",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "MessagePart",
    "UserMessagePart",
    "AssistantMessagePart",
    "UserMessage",
    "ObjectType",
    "AssistantMessage",
    "ToolMessage",
    "Message",
    "NonSystemMessage",
    "Prompt",
    "ResponseFormatText",
    "ResponseFormatJson",
    "ResponseFormat",
]



from . import Plugin, tool
from typing import Annotated


class Calculator(Plugin):
    @tool
    def evaluate(
        self,
        expression: Annotated[
            str, "The math expression to evaluate. Must be an valid python expression."
        ],
    ):
        """Execute a math expression and return the result. The expression must be an valid python expression that can be execuated by `eval()`."""
        result = eval(expression)
        return result



import base64
from collections.abc import AsyncGenerator
from datetime import datetime
import json
import os
from typing import Any, Literal, Sequence, override
from uuid import uuid4
import httpx
from pydantic import HttpUrl

from agentia.tools.tools import ToolSet
from . import LLMOptions, GenerationResult, Provider

from ...spec import *
from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
)
import openai
from openai.types import CompletionUsage


def _gen_id() -> str:
    return uuid4().hex


class OpenAIAPIProvider(Provider):
    def __init__(
        self,
        provider: str,
        model: str,
        api_key: str | None = None,
        base_url: str | None = None,
    ):
        if model.endswith(":think"):
            model = model[: -len(":think")]
            think = True
        else:
            think = False
        super().__init__(provider=provider, model=model)
        api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable not set")
        if "OPENAI_BASE_URL" in os.environ:
            base_url = os.environ["OPENAI_BASE_URL"]
        self.api_key = api_key
        self.base_url = base_url
        self.extra_headers: dict[str, str] = {}
        self.extra_body: dict[str, Any] = {}
        if think:
            self.reasoning = True
            self.enable_reasoning()
        else:
            self.reasoning = False

    def client(self, client: httpx.AsyncClient) -> openai.AsyncOpenAI:
        return openai.AsyncOpenAI(
            api_key=self.api_key, base_url=self.base_url, http_client=client
        )

    def enable_reasoning(self) -> None:
        self.extra_body["reasoning"] = {
            "enabled": True,
            "effort": "high",
        }

    def _to_oai_content_part(
        self, p: MessagePart, index: int
    ) -> ChatCompletionContentPartParam:
        match p.type:
            case "text":
                return ChatCompletionContentPartTextParam(type="text", text=p.text)
            case "file":
                if p.media_type.startswith("image/"):
                    if isinstance(p.data, str):
                        if p.data.startswith("data:"):
                            # this is a data URL
                            url = p.data
                        elif p.data.startswith("http://") or p.data.startswith(
                            "https://"
                        ):
                            # this is a URL
                            url = p.data
                        else:
                            # this is a base64 encoded string
                            url = f"data:{p.media_type};base64,{p.data}"
                    elif isinstance(p.data, HttpUrl):
                        url = str(p.data)
                    else:
                        assert isinstance(p.data, bytes)
                        base64_data = base64.b64encode(p.data).decode(encoding="utf-8")
                        url = f"data:{p.media_type};base64,{base64_data}"
                    detail = (p.provider_options or {}).get("imageDetail", None)
                    if detail not in ["auto", "low", "high"]:
                        detail = "auto"
                    return ChatCompletionContentPartImageParam(
                        type="image_url",
                        image_url={"url": url, "detail": detail},  # type: ignore
                    )
                if p.media_type.startswith("video/"):
                    if isinstance(p.data, str):
                        if p.data.startswith("data:"):
                            # this is a data URL
                            url = p.data
                        elif p.data.startswith("http://") or p.data.startswith(
                            "https://"
                        ):
                            # this is a URL
                            url = p.data
                        else:
                            # this is a base64 encoded string
                            url = f"data:{p.media_type};base64,{p.data}"
                    elif isinstance(p.data, HttpUrl):
                        url = str(p.data)
                    else:
                        assert isinstance(p.data, bytes)
                        base64_data = base64.b64encode(p.data).decode(encoding="utf-8")
                        url = f"data:{p.media_type};base64,{base64_data}"
                    return {"type": "video_url", "video_url": {"url": url}}  # type: ignore
                elif p.media_type.startswith("audio/"):
                    if isinstance(p.data, HttpUrl) or (
                        isinstance(p.data, str)
                        and (
                            p.data.startswith("http://")
                            or p.data.startswith("https://")
                        )
                    ):
                        raise ValueError("audio file parts with URLs")
                    if isinstance(p.data, str):
                        data = p.data
                    else:
                        assert isinstance(p.data, bytes)
                        data = p.data.decode(encoding="utf-8")
                    format: Literal["wav", "mp3"]
                    if p.media_type == "audio/wav":
                        format = "wav"
                    elif p.media_type == "audio/mpeg" or p.media_type == "audio/mp3":
                        format = "mp3"
                    else:
                        raise ValueError(
                            f"audio file parts with media type {p.media_type}"
                        )
                    return ChatCompletionContentPartInputAudioParam(
                        type="input_audio", input_audio={"data": data, "format": format}
                    )
                else:
                    if isinstance(p.data, str) and p.data.startswith("file-"):
                        # this is a file ID
                        return {"type": "file", "file": {"file_id": p.data}}
                    else:
                        url = None
                        if isinstance(p.data, str):
                            if p.data.startswith("data:"):
                                url = p.data
                            else:
                                # this is a base64 encoded string
                                url = f"data:{p.media_type};base64,{p.data}"
                        elif isinstance(p.data, HttpUrl):
                            url = str(p.data)
                        else:
                            assert isinstance(p.data, bytes)
                            base64_data = base64.b64encode(p.data).decode(
                                encoding="utf-8"
                            )
                            url = f"data:{p.media_type};base64,{base64_data}"
                        return {
                            "type": "file",
                            "file": {
                                "filename": p.filename or f"file-{index}",
                                "file_data": url,
                            },
                        }
            case _:
                raise ValueError(f"content part of type {p.type}")

    def _to_oai_messages(self, prompt: Prompt) -> list[ChatCompletionMessageParam]:
        r: list[ChatCompletionMessageParam] = []
        for m in prompt:
            if m.role == "system":
                r.append(
                    ChatCompletionSystemMessageParam(role="system", content=m.content)
                )
            elif m.role == "user":
                content = m.content_list
                if len(content) == 1 and content[0].type == "text":
                    r.append(
                        ChatCompletionUserMessageParam(
                            role="user", content=content[0].text
                        )
                    )
                    continue
                parts: list[ChatCompletionContentPartParam] = []
                for i, p in enumerate(content):
                    parts.append(self._to_oai_content_part(p, i))
                r.append(ChatCompletionUserMessageParam(role="user", content=parts))
            elif m.role == "assistant":
                text = ""
                reasoning = ""
                images: list[Any] = []
                tool_calls: list[ChatCompletionMessageToolCallParam] = []
                for i, p in enumerate(m.content_list):
                    if p.type == "text":
                        text += p.text
                    elif p.type == "reasoning":
                        reasoning += p.text
                    elif p.type == "file":
                        if p.media_type.startswith("image/"):
                            images.append(
                                {"type": "image_url", "image_url": {"url": p.to_url()}}
                            )
                        else:
                            raise ValueError(
                                f"Only image files are supported in assistant messages, got media type {p.media_type}"
                            )
                    elif p.type == "tool-call":
                        tool_calls.append(
                            {
                                "id": p.tool_call_id,
                                "type": "function",
                                "function": {
                                    "name": p.tool_name,
                                    "arguments": json.dumps(p.input),
                                },
                            }
                        )
                p = ChatCompletionAssistantMessageParam(
                    role="assistant",
                    content=text if text else None,
                    tool_calls=tool_calls,
                )
                if reasoning:
                    p["reasoning"] = reasoning  # type: ignore
                if images:
                    p["images"] = images  # type: ignore
                r.append(p)
            elif m.role == "tool":
                for p in m.content:
                    val = json.dumps(p.output)
                    r.append(
                        ChatCompletionToolMessageParam(
                            role="tool", tool_call_id=p.tool_call_id, content=val
                        )
                    )
            else:
                raise ValueError(f"Unsupported message role {m.role}")
        return r

    def _get_finish_reason(
        self,
        choice: Literal[
            "stop", "length", "tool_calls", "content_filter", "function_call"
        ],
    ) -> FinishReason:
        match choice:
            case "stop":
                return "stop"
            case "length":
                return "length"
            case "tool_calls" | "function_call":
                return "tool-calls"
            case "content_filter":
                return "content-filter"
            case _:
                return "unknown"

    def _prepare_tools(
        self,
        tools: Sequence[Tool] | None,
        tool_choice: ToolChoice | None,
    ) -> tuple[
        list[ChatCompletionToolParam] | None, ChatCompletionToolChoiceOptionParam | None
    ]:
        oai_tools: list[ChatCompletionToolParam] = []
        if not tools:
            return None, None
        for t in tools:
            match t.type:
                case "function":
                    f = {
                        "name": t.name,
                        "description": t.description,
                        "parameters": t.input_schema,  # type: ignore
                    }
                    for k in list(f.keys()):
                        if f[k] is None:
                            del f[k]
                    oai_tools.append(
                        {
                            "type": "function",
                            "function": f,  # type: ignore
                        }
                    )
        if not tool_choice:
            return oai_tools, None
        match tool_choice:
            case "none":
                return oai_tools, "none"
            case "auto":
                return oai_tools, "auto"
            case "required":
                return oai_tools, "required"
            case _:
                assert isinstance(tool_choice, str)
                tc: ChatCompletionToolChoiceOptionParam = {
                    "type": "function",
                    "function": {"name": tool_choice},
                }
                return oai_tools, tc

    def _get_usage(self, u: CompletionUsage | None) -> Usage:
        if not u:
            return Usage()
        return Usage(
            input_tokens=u.prompt_tokens,
            output_tokens=u.completion_tokens,
            total_tokens=u.total_tokens,
            reasoning_tokens=(
                u.completion_tokens_details.reasoning_tokens
                if u.completion_tokens_details
                else None
            ),
            cached_input_tokens=(
                u.prompt_tokens_details.cached_tokens
                if u.prompt_tokens_details
                else None
            ),
        )

    def _prepare_args(
        self, prompt: Prompt, tool_set: ToolSet, options: LLMOptions
    ) -> dict[str, Any]:
        msgs = self._to_oai_messages(prompt)
        rf = options.response_format
        if not rf or rf.type == "text":
            response_format = None
        elif not rf.json_schema:
            response_format = {"type": "json_object"}
        else:
            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "schema": rf.json_schema,
                    "strict": True,
                    "name": rf.name or "response",
                    "description": rf.description,
                },
            }
        schema = tool_set.get_schema()
        tools, tool_choice = self._prepare_tools(schema, options.tool_choice)
        args: dict[str, Any] = {
            "model": self.model,
            "messages": msgs,
            "temperature": options.temperature,
            "top_p": options.top_p,
            "max_tokens": options.max_output_tokens,
            "frequency_penalty": options.frequency_penalty,
            "presence_penalty": options.presence_penalty,
            "response_format": response_format,
            "stop": options.stop_sequences,
            "seed": options.seed,
            "tools": tools,
            "tool_choice": tool_choice,
        }
        for k in list(args.keys()):
            if args[k] is None:
                del args[k]
        return args

    def _get_extra_body(self, options: LLMOptions) -> dict[str, Any]:
        body: dict[str, Any] = {}
        if options.provider_options:
            body.update(options.provider_options)
        return body

    @override
    async def do_generate(
        self,
        prompt: Prompt,
        tool_set: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> GenerationResult:
        args = self._prepare_args(prompt, tool_set, options)
        if t := os.environ.get("AGENTIA_TIMEOUT"):
            timeout = float(t)
        else:
            timeout = None
        response = await self.client(client).chat.completions.create(
            **args,
            extra_headers=self.extra_headers,
            extra_body=self._get_extra_body(options),
            stream=False,
            timeout=timeout,
        )
        choice = response.choices[0]
        parts: Sequence[AssistantMessagePart] = []
        tool_calls: list[ToolCall] = []

        if self.reasoning and hasattr(choice.message, "reasoning"):
            reasoning_text: str = choice.message.reasoning  # type: ignore
            if reasoning_text and reasoning_text.strip() != "":
                parts.append(MessagePartReasoning(text=reasoning_text))

        text = choice.message.content
        if text:
            parts.append(MessagePartText(text=text))
        for tc in choice.message.tool_calls or []:
            assert tc.type == "function"
            args = json.loads(tc.function.arguments) if tc.function.arguments else {}
            t = ToolCall(
                tool_call_id=tc.id or _gen_id(), tool_name=tc.function.name, input=args
            )
            parts.append(
                MessagePartToolCall(
                    tool_call_id=t.tool_call_id, tool_name=t.tool_name, input=t.input
                )
            )
            tool_calls.append(t)

        annotations: list[Annotation] | None = None
        for a in choice.message.annotations or []:
            if not annotations:
                annotations = []
            annotations.append(
                Annotation(
                    title=a.url_citation.title,
                    url=a.url_citation.url,
                    start=a.url_citation.start_index,
                    end=a.url_citation.end_index,
                )
            )

        return GenerationResult(
            message=AssistantMessage(content=parts, annotations=annotations),
            tool_calls=tool_calls,
            finish_reason=self._get_finish_reason(choice.finish_reason),
            usage=self._get_usage(response.usage),
            provider_metadata=None,
        )

    @override
    async def do_stream(
        self,
        prompt: Prompt,
        tool_set: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> AsyncGenerator[StreamPart, None]:
        args = self._prepare_args(prompt, tool_set, options)
        if t := os.environ.get("AGENTIA_TIMEOUT"):
            timeout = float(t)
        else:
            timeout = None
        response = await self.client(client).chat.completions.create(
            **args,
            extra_headers=self.extra_headers,
            extra_body=self._get_extra_body(options),
            stream=True,
            stream_options={
                "include_usage": True,
            },
            timeout=timeout,
        )
        started = False
        streaming_text = False
        streaming_reasoning = False
        tool_calls: list[ToolCall | None] = []
        finished = False
        async for chunk in response:
            if finished:
                continue
            if not started:
                started = True
                yield StreamPartStreamStart(
                    id=_gen_id(),
                    timestamp=datetime.fromtimestamp(chunk.created),
                    model_id=self.model,
                )
            if not chunk.choices:
                yield StreamPartStreamEnd(
                    usage=self._get_usage(chunk.usage), finish_reason="stop"
                )
                continue

            choice = chunk.choices[0]
            delta = choice.delta

            if rd := getattr(delta, "reasoning", None):
                if not self.reasoning:
                    continue
                if not streaming_reasoning:
                    streaming_reasoning = True
                    yield StreamPartReasoningStart(id=_gen_id())
                yield StreamPartReasoningDelta(id=_gen_id(), delta=rd)
            elif streaming_reasoning:
                streaming_reasoning = False
                yield StreamPartReasoningEnd(id=_gen_id())

            if delta.content:
                if not streaming_text:
                    streaming_text = True
                    yield StreamPartTextStart(id=_gen_id())
                yield StreamPartTextDelta(id=_gen_id(), delta=delta.content)
            elif streaming_text:
                streaming_text = False
                yield StreamPartTextEnd(id=_gen_id())

            if delta.tool_calls:
                for tc in delta.tool_calls:
                    index = tc.index
                    if index >= len(tool_calls) or tool_calls[index] is None:
                        tool_calls.extend([None] * (index - len(tool_calls) + 1))
                        assert tc.id and tc.function and tc.function.name
                        tool_calls[index] = ToolCall(
                            tool_call_id=tc.id, tool_name=tc.function.name, input=""
                        )
                    tc_obj = tool_calls[index]
                    assert tc_obj is not None
                    if tc.function and tc.function.arguments:
                        assert isinstance(tc_obj.input, str)
                        tc_obj.input += tc.function.arguments or ""

            if choice.finish_reason:
                for i, tc in enumerate(tool_calls):
                    if not tc:
                        continue
                    assert isinstance(tc.input, str)
                    tc.input = json.loads(tc.input or "{}")
                    yield tc
                yield StreamPartStreamEnd(
                    usage=self._get_usage(chunk.usage),
                    finish_reason=self._get_finish_reason(choice.finish_reason),
                )
                finished = True



from . import Plugin, tool
from typing import Annotated, Literal
import os
from tavily import AsyncTavilyClient


class Search(Plugin):
    def __init__(
        self,
        tavily_api_key: str | None = None,
    ):
        super().__init__()
        api_key = tavily_api_key or os.environ.get("TAVILY_API_KEY")
        if not api_key:
            raise ValueError("TAVILY_API_KEY is required for SearchPlugin")
        self.__tavily = AsyncTavilyClient(api_key=api_key)

    @tool
    async def web_search(
        self,
        query: Annotated[
            str, "The search query. Please be as specific and verbose as possible."
        ],
        topic: Annotated[
            Literal["general", "news", "finance"] | None,
            "The topic of the search. Default is 'general'.",
        ] = None,
        time_range: Annotated[
            Literal["day", "week", "month", "year"] | None,
            "The time range back from the current date based on publish date or last updated date.",
        ] = None,
        max_results: Annotated[
            int | None,
            "The maximum number of search results to return. Default is 5. It must be between 0 and 20.",
        ] = None,
    ):
        """
        Perform web search on the given query.
        Returning the top related search results in json format.
        When necessary, you need to combine this tool with the get_webpage_content tools (if available), to browse the web in depth by jumping through links.
        """

        tavily_results = await self.__tavily.search(
            query=query,
            search_depth="advanced",
            # max_results=10,
            include_answer=True,
            include_images=True,
            include_image_descriptions=True,
            topic=topic,  # type: ignore
            time_range=time_range,  # type: ignore
            max_results=max_results,  # type: ignore
        )
        return tavily_results


_ALL_COUNTRIES = [
    # Default country
    "Australia",
    # Other countries
    "Afghanistan",
    "Albania",
    "Antarctica",
    "Algeria",
    "American Samoa",
    "Andorra",
    "Angola",
    "Antigua and Barbuda",
    "Azerbaijan",
    "Argentina",
    "Austria",
    "The Bahamas",
    "Bahrain",
    "Bangladesh",
    "Armenia",
    "Barbados",
    "Belgium",
    "Bhutan",
    "Bolivia",
    "Bosnia and Herzegovina",
    "Botswana",
    "Brazil",
    "Belize",
    "Solomon Islands",
    "Brunei",
    "Bulgaria",
    "Myanmar (Burma)",
    "Burundi",
    "Cambodia",
    "Cameroon",
    "Canada",
    "Cabo Verde",
    "Central African Republic",
    "Sri Lanka",
    "Chad",
    "Chile",
    "China",
    "Christmas Island",
    "Cocos (Keeling) Islands",
    "Colombia",
    "Comoros",
    "Republic of the Congo",
    "Democratic Republic of the Congo",
    "Cook Islands",
    "Costa Rica",
    "Croatia",
    "Cyprus",
    "Czechia",
    "Benin",
    "Denmark",
    "Dominica",
    "Dominican Republic",
    "Ecuador",
    "El Salvador",
    "Equatorial Guinea",
    "Ethiopia",
    "Eritrea",
    "Estonia",
    "South Georgia and the South Sandwich Islands",
    "Fiji",
    "Finland",
    "France",
    "French Polynesia",
    "French Southern and Antarctic Lands",
    "Djibouti",
    "Gabon",
    "Georgia",
    "The Gambia",
    "Germany",
    "Ghana",
    "Kiribati",
    "Greece",
    "Grenada",
    "Guam",
    "Guatemala",
    "Guinea",
    "Guyana",
    "Haiti",
    "Heard Island and McDonald Islands",
    "Vatican City",
    "Honduras",
    "Hungary",
    "Iceland",
    "India",
    "Indonesia",
    "Iraq",
    "Ireland",
    "Israel",
    "Italy",
    "Cote d'Ivoire",
    "Jamaica",
    "Japan",
    "Kazakhstan",
    "Jordan",
    "Kenya",
    "South Korea",
    "Kuwait",
    "Kyrgyzstan",
    "Laos",
    "Lebanon",
    "Lesotho",
    "Latvia",
    "Liberia",
    "Libya",
    "Liechtenstein",
    "Lithuania",
    "Luxembourg",
    "Madagascar",
    "Malawi",
    "Malaysia",
    "Maldives",
    "Mali",
    "Malta",
    "Mauritania",
    "Mauritius",
    "Mexico",
    "Monaco",
    "Mongolia",
    "Moldova",
    "Montenegro",
    "Morocco",
    "Mozambique",
    "Oman",
    "Namibia",
    "Nauru",
    "Nepal",
    "Netherlands",
    "Curacao",
    "Sint Maarten",
    "Caribbean Netherlands",
    "New Caledonia",
    "Vanuatu",
    "New Zealand",
    "Nicaragua",
    "Niger",
    "Nigeria",
    "Niue",
    "Norfolk Island",
    "Norway",
    "Northern Mariana Islands",
    "United States Minor Outlying Islands",
    "Micronesia",
    "Marshall Islands",
    "Palau",
    "Pakistan",
    "Panama",
    "Papua New Guinea",
    "Paraguay",
    "Peru",
    "Philippines",
    "Pitcairn Islands",
    "Poland",
    "Portugal",
    "Guinea-Bissau",
    "Timor-Leste",
    "Qatar",
    "Romania",
    "Rwanda",
    "Saint Helena, Ascension and Tristan da Cunha",
    "Saint Kitts and Nevis",
    "Saint Lucia",
    "Saint Pierre and Miquelon",
    "Saint Vincent and the Grenadines",
    "San Marino",
    "Sao Tome and Principe",
    "Saudi Arabia",
    "Senegal",
    "Serbia",
    "Seychelles",
    "Sierra Leone",
    "Singapore",
    "Slovakia",
    "Vietnam",
    "Slovenia",
    "Somalia",
    "South Africa",
    "Zimbabwe",
    "Spain",
    "Suriname",
    "Eswatini",
    "Sweden",
    "Switzerland",
    "Tajikistan",
    "Thailand",
    "Togo",
    "Tokelau",
    "Tonga",
    "Trinidad and Tobago",
    "United Arab Emirates",
    "Tunisia",
    "Turkiye",
    "Turkmenistan",
    "Tuvalu",
    "Uganda",
    "Ukraine",
    "North Macedonia",
    "Egypt",
    "United Kingdom",
    "Guernsey",
    "Jersey",
    "Isle of Man",
    "Tanzania",
    "United States",
    "Burkina Faso",
    "Uruguay",
    "Uzbekistan",
    "Venezuela",
    "Wallis and Futuna",
    "Samoa",
    "Yemen",
    "Zambia",
]
